/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <fcntl.h>
#ifdef	MALLOC_H
#include <malloc.h>
#endif
#ifdef	STDLIB_H
#include <stdlib.h>
#endif
#include "is_types.h"
#include "is_proto.h"
#include "isam.h"

#ifdef BSD
#include <strings.h>
#define memcpy(a, b, c) bcopy((b), (a), (c))
#else
#include <string.h>
#define	bzero(a, b)	memset ((a), 0, (b))
#endif /* BSD */

#ifndef lint
static	char	sccsid[] = "@(#)db_open.c	1.10	10:32:33	2/26/92";
#endif

extern	schema_t dict_schema;
extern	is_index_t dict_index;
extern	int	nbuffers;

#if !defined(STLIB_H)
extern	char	*getenv ();
#endif
#if !defined(MALLOC_H) && !defined(STDLIB_H)
extern	char	*malloc();
#endif

extern	int	db_errno;

int
db_open_index (dbm, filenum, name)
dbm_t	*dbm;
int	filenum;
char	*name;
{
	file_t	*file;
	is_index_t	*index;
	int	i;

	if (filenum < 0 || filenum >= dbm->d_file_count ||
			dbm->d_files[filenum].f_file_name[0] == '\0')
		return (db_errno = DB_INVALID), -1;

	file = &dbm->d_files[filenum];
	if (file->f_schema->s_fields == 0)
		return (db_errno = DB_NOTOPEN), -1;

	if (! (index = file->f_index))
		return (db_errno = DB_NO_FIELD), -1;

	if (name == 0) {

		/*
		 * Use the primary index.
		 */

		file->f_current_index = file->f_key_index = 1;
		file->f_key_block = 0;
		file->f_key_slot = 0;

		return 1;
	}

	/*
	 * Search for an index with the same name as "name".  It
	 * will be made the current index if it is found.
	 */

	for (i = 1;i < file->f_schema->s_field_count;i++) {
		if (file->f_schema->s_fields[i].f_type != T_INDEX)
			continue;

		if (strncmp (name, file->f_schema->s_fields[i].f_name,
				MAXFIELDNAME) == 0)
			break;
	}
	if (i == file->f_schema->s_field_count)
		return (db_errno = DB_NO_FIELD), -1;

	file->f_current_index = file->f_key_index =
		file->f_schema->s_fields[i].f_offset;
	file->f_key_block = 0;
	file->f_key_slot = 0;

	return file->f_current_index;
}

int
db_open_field (dbm, filenum, name)
dbm_t	*dbm;
int	filenum;
char	*name;
{
	file_t	*file;
	int	i;

	if (filenum < 0 || filenum >= dbm->d_file_count ||
			dbm->d_files[filenum].f_file_name[0] == '\0')
		return (db_errno = DB_INVALID), -1;

	file = &dbm->d_files[filenum];
	if (file->f_schema->s_fields == 0)
		return (db_errno = DB_NOTOPEN), -1;

	for (i = 1;i < file->f_schema->s_field_count;i++) {
		if (strncmp (name, file->f_schema->s_fields[i].f_name,
				MAXFIELDNAME) == 0)
			break;
	}
	if (i == file->f_schema->s_field_count)
		return (db_errno = DB_NO_FIELD), -1;

	return i;
}

/*
 * db_open_file - open an existing database file
 */

int
db_open_file (dbm, name)
dbm_t	*dbm;
char	*name;
{
	file_t	*dict;
	file_t	*file;
	schema_t *schema;
	field_t	*fields;
	is_index_t *index;
	sort_key_t *keys;
	dict_t	field;
	char	filename[MAXFILENAME];
	char	fieldname[MAXFIELDNAME];
	int	file_keys[2];
	int	i, j, k;
	off_t	record;
	int	size;

	dict = &dbm->d_dictionary;

	/*
	 * Scan for the name in the dictionary to see if the file is
	 * part of this database.  If the name isn't found, this file
	 * is not a member of the database.
	 */

	if (dbm->d_files) {
		for (i = 0;i < dbm->d_file_count;i++) {
			if (! dbm->d_files[i].f_file_name[0])
				continue;

			if (strncmp (dbm->d_files[i].f_file_name,
						name, MAXFILENAME) == 0) {
				if (dbm->d_files[i].f_schema->s_fields)
					return i;
				else
					break;
			}
		}
	}
	if (i == dbm->d_file_count)
		return (db_errno = DB_NO_FILE), -1;

	bzero (filename, sizeof filename);
	strncpy (filename, name, sizeof filename);

	if (__db_read_record (dict, 3, filename, IS_EQUAL, &field) == -1)
		return (db_errno = DB_BAD_DB), -1;

	if ((i = field.f_fileno) < 0 || i >= dbm->d_file_count)
		return (db_errno = DB_BAD_DB), -1;

	/* 
	 * Open the ISAM file and perform some simple checks on
	 * the file structure.  This binary must be able to process
	 * the ISAM file that was just opened, given the magic
	 * numbers and block sizes.
	 */

	if ((file = isam_file_open (name, dbm->d_mode)) == 0)
		return -1;

	if (file->f_magic && file->f_magic != FILEMAGIC) {
		db_errno = DB_BAD_DB;
		goto fail;
	}
	if (file->f_bsize && file->f_bsize > MAXBLOCKSIZE) {
		db_errno = DB_BAD_DB;
		goto fail;
	}
	if (! (file->f_schema = dbm->d_files[i].f_schema)) {
		db_errno = DB_BAD_DB;
		goto fail;
	}
	dbm->d_files[i] = *file;
	file = &dbm->d_files[i];

	schema = file->f_schema;
	if (! (schema->s_fields = (field_t *)
			malloc (schema->s_field_count * sizeof (field_t)))) {
		db_errno = DB_NOMEM;
		goto fail;
	}
	fields = schema->s_fields;
	bzero (fields, schema->s_field_count * sizeof (field_t));

	/*
	 * The file number in the key is set to the position of the
	 * file in the dictionary.  The field number in the key is
	 * varied up to the total number of fields.  As each record
	 * is read in using this key, the schema for the file is
	 * gradually built up.
	 */

	file_keys[0] = i;
	for (size = 0, i = 1;i < schema->s_field_count;i++) {
		file_keys[1] = i;
		if (__db_read_record (dict, 2, file_keys,
				IS_LESSEQ, &field) == -1) {
			db_errno = DB_BAD_DB;
			goto fail;
		}
		fields[i].f_type = field.f_fieldtype;
		fields[i].f_size = field.f_fieldsize;
		fields[i].f_offset = size;
		strncpy (fields[i].f_name, field.f_fieldname, MAXFIELDNAME);

		if (field.f_fieldtype == T_INDEX) {
			if (file->f_index)
				file->f_index->i_key_count++;
			else {
				if (! (file->f_index = (is_index_t *)
						malloc (sizeof (is_index_t))))
					return (db_errno = DB_NOMEM), -1;

				file->f_index->i_key_count = 1;
				file->f_index->i_keys = 0;
			}
			fields[i].f_offset = file->f_index->i_key_count;
		} else {
			size += field.f_fieldsize;
		}
	}
	if (index = file->f_index) {
		if (! (index->i_keys = (sort_key_t *)
				malloc ((file->f_index->i_key_count + 1) *
					sizeof (sort_key_t)))) {
			db_errno = DB_NOMEM;
			goto fail;
		}
		for (i = 1;i < schema->s_field_count;i++) {
			if (fields[i].f_type != T_INDEX)
				continue;

			file_keys[1] = i;
			if (__db_read_record (dict, 2, file_keys, IS_EQUAL,
					&field) == -1) {
				db_errno = DB_BAD_DB;
				goto fail;
			}
			if (fields[i].f_offset > index->i_key_count) {
				db_errno = DB_BAD_DB;
				goto fail;
			}
			keys = &index->i_keys[fields[i].f_offset];
			keys->k_field_number = i;
			if ((keys->k_field_count = field.f_fieldsize) > 0) {
				keys->k_unique = 1;
			} else {
				keys->k_field_count = - keys->k_field_count;
				keys->k_unique = 0;
			}
			memcpy (keys->k_fields, field.f_fieldkeys,
				keys->k_field_count * sizeof (short));

			for (keys->k_key_size = 0, j = 0;
					j < keys->k_field_count;j++) {
				k = keys->k_fields[j] & ~ IS_DESCENDING;
				keys->k_key_size += fields[k].f_size;
			}
			if (keys->k_unique == 0) {
				keys->k_field_count++;
				keys->k_key_size += sizeof (off_t);
			}
		}
		file->f_current_index = 1;
	} else {
		file->f_index = 0;
	}

	/*
	 * It is assumed that MAXRECSIZE is the largest a record
	 * will ever be.
	 */

	if (size > MAXRECSIZE) {
		db_errno = DB_BAD_DB;
		goto fail;
	}
	file->f_record_size = size;
	if (! (file->f_data_buffer = (char *) malloc (size))) {
		db_errno = DB_NOMEM;
		goto fail;
	}

	/*
	 * Every file that is open gets 4 ISAM buffers.  They are all
	 * shared, so it helps when plenty of files are open.
	 */

	for (j = i = 0;i < dbm->d_file_count;i++) {
		if (! dbm->d_files[i].f_schema)
			continue;

		if (dbm->d_files[i].f_schema->s_fields)
			j++;
	}
	if (j * 4 > nbuffers)
		(void) isam_setbuf (j * 4);

	return file_keys[0];

fail:

	/* 
	 * See how far along the file open has proceeded.  Once the
	 * schema has been started on, the file must be closed with
	 * db_close_file, otherwise the ISAM close routine is enough.
	 */

	if (file->f_schema->s_fields)
		__db_close_file (dbm, file);
	else
		isam_close (file);

	return -1;
}

/*
 * db_open - open an existing database or create a new one
 *
 * db_open() opens existing databases and loads the file lists from the
 * data dictionary, or creates a new database with a completely empty
 * data dictionary if the named database does not exist.
 */

dbm_t *
db_open (name, mode)
char	*name;
int	mode;
{
	dbm_t	dbm;
	dbm_t	*new_dbm;
	file_t	*dict;
	file_t	*file;
	schema_t *schema;
	is_index_t *index;
	off_t	record;
	dict_t	field;
	int	file_keys[2];
	int	i;
	char	*cp;

	bzero (&dbm, sizeof dbm);

	/*
	 * Attempt to open the data dictionary.  If the mode is O_RDWR, we
	 * may try to re-open it is O_RDONLY if the open fails.  This is
	 * because a user may try to open a database for R/W when they don't
	 * have write access to the dictionary.
	 */

	if ((dict = isam_file_open (name, mode)) == 0) {
		if (mode != O_RDWR ||
				! (dict = isam_file_open (name, O_RDONLY)))
			return 0;
	}

	/*
	 * New data dictionaries need the correct magic number inserted
	 * in the file header.  This is because the isam_file_open routine
	 * doesn't know it is creating a dictionary.
	 */

	if (mode & O_CREAT) {
		header_t header;

		if (lseek (dict->f_key_fd, 0L, 0) != 0) {
			db_errno = DB_EIO;
			goto fail;
		}
		if (read (dict->f_key_fd, &header, sizeof header) !=
				sizeof header) {
			db_errno = DB_EIO;
			goto fail;
		}
		dict->f_magic = header.h_magic = DICTMAGIC;

		if (lseek (dict->f_key_fd, 0L, 0) != 0) {
			db_errno = DB_EIO;
			goto fail;
		}
		if (write (dict->f_key_fd, &header, sizeof header) !=
				sizeof header) {
			db_errno = DB_EIO;
			goto fail;
		}
	}

	/*
	 * Check the magic numbers and block sizes for correctness
	 */

	if (dict->f_magic && dict->f_magic != DICTMAGIC) {
		db_errno = DB_BAD_DB;
		goto fail;
	}
	if (dict->f_bsize && dict->f_bsize > MAXBLOCKSIZE) {
		db_errno = DB_BAD_DB;
		goto fail;
	}

	/*
	 * Set up the schema for the data dictionary.  It is hardcoded
	 * in the library.
	 */

	dict->f_record_size = sizeof field;
	if (! (dict->f_data_buffer = malloc (dict->f_record_size))) {
		db_errno = DB_NOMEM;
		goto fail;
	}
	dict->f_schema = &dict_schema;
	dict->f_index = &dict_index;
	if ((dict->f_ticket = isam_get_ticket (dict)) == -1)
		goto fail;

	dbm.d_dictionary = *dict;
	dbm.d_file_count = 0;
	dbm.d_files = 0;
	dbm.d_mode = mode;

	if (mode & O_CREAT) {
		if (! (new_dbm = (dbm_t *) malloc (sizeof dbm))) {
			db_errno = DB_NOMEM;
			goto fail;
		}
		*new_dbm = dbm;
		return new_dbm;
	}

	/*
	 * Figure out how many files there are in this database.  We look
	 * for the first file with a file number less than 9999.  This will
	 * be the highest numbered file in the database.  Add one to that
	 * and you get the number of files in the database.
	 */

	file_keys[0] = 9999;
	if (__db_read_record (dict, 1, &file_keys[0], IS_LESSEQ, &field) == -1)
		goto fail;

	dbm.d_file_count = field.f_fileno + 1;
	if ((dbm.d_files = (file_t *) malloc (dbm.d_file_count *
			sizeof (file_t))) == 0) {
		db_errno = DB_NOMEM;
		goto fail;
	}
	bzero (dbm.d_files, dbm.d_file_count * sizeof (file_t));

	/*
	 * Now that I know how many files I have, I can run off and see
	 * how many fields each file has.
	 */

	for (file_keys[0] = 0;file_keys[0] < dbm.d_file_count;file_keys[0]++) {

		/*
		 * I use the same mechanism as before - I ask for the highest
		 * numbered field in the file.  This gives me the largest
		 * numbered field, and one more than that is the total number
		 * of fields in this database file.
		 */

		file_keys[1] = 9999;
		file = &dbm.d_files[file_keys[0]];
		if (__db_read_record (dict, 2, file_keys,
				IS_LESSEQ, &field) == -1) {
			memset (file, 0, sizeof *file);
			continue;
		}

		/*
		 * The LESSEQ comparision might have given me the previously
		 * numbered file - skip this record if that's the case.  This
		 * can happen if a file was deleted and it's old number not
		 * reused yet.
		 */

		if (field.f_fileno != file_keys[0]) {
			memset (file, 0, sizeof *file);
			continue;
		}

		/*
		 * Make up the file entry for this database file.
		 */

		if ((schema = (schema_t *) malloc (sizeof *schema)) == 0) {
			db_errno = DB_NOMEM;
			goto fail;
		}
		file->f_schema = schema;
		file->f_data_fd = -1;
		file->f_key_fd = -1;
		strncpy (file->f_file_name, field.f_fieldfile, MAXFILENAME);

		schema->s_field_count = field.f_fieldno + 1;
		schema->s_fields = 0;
	}
	if (! (new_dbm = (dbm_t *) malloc (sizeof dbm))) {
		db_errno = DB_NOMEM;
		goto fail;
	}
	*new_dbm = dbm;

	if (cp = getenv ("DBBUFFERS"))
		(void) isam_setbuf (atoi (cp));

	return new_dbm;

fail:

	/*
	 * Close the open file descriptors for the data dictionary.
	 */

	if (dict->f_key_fd != -1)
		close (dict->f_key_fd);

	if (dict->f_data_fd != -1)
		close (dict->f_data_fd);

	/*
	 * Free up the allocated memory.
	 */

	if (dict->f_data_buffer)
		free (dict->f_data_buffer);

	if (dbm.d_files) {
		for (i = 0;i < dbm.d_file_count;i++)
			if (dbm.d_files[i].f_schema)
				free (dbm.d_files[i].f_schema);

		free (dbm.d_files);
	}
	return 0;
}
