/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>
#include "is_types.h"
#include "is_proto.h"
#include "patchlevel.h"

#ifdef BSD
#define	memcpy(d,s,n)	bcopy(s,d,n)
#endif /* BSD */

#ifndef lint
static	char	sccsid[] = "@(#)dbck.c	1.5	10:32:45	2/26/92";
#endif

dbm_t	*db_open();
int	db_open_file();
int	db_open_index();
long	isam_next();
char	*malloc();

long	max_records;
long	record_count;
long	max_index;
long	index_count;
char	*record_bits;
char	*index_blocks;

extern	int	db_errno;

int
test_record_bit (record)
off_t	record;
{
	if (record < 0 || record >= max_records)
		return 0;

	return record_bits[record / 8] & (1 << (record & 07));
}

int
set_record_bit (record)
off_t	record;
{
	if (record < 0 || record >= max_records) {
		printf ("illegal record %d\n", record);
		return -1;
	}
	if (record_bits[record / 8] & (1 << (record & 07))) {
		printf ("duplicate record %d\n", record);
		return -1;
	}
	record_bits[record / 8] |= (1 << (record & 07));
	return 0;
}

int
clear_record_bit (record)
off_t	record;
{
	if (record < 0 || record >= max_records) {
		printf ("illegal record %d\n", record);
		return -1;
	}
	record_bits[record / 8] &= ~(1 << (record & 07));
	return 0;
}

int
set_index_block (block, index)
off_t	block;
int	index;
{
	if (block < 0 || block >= max_index) {
		printf ("illegal index block %d\n", block);
		return -1;
	}
	if (index_blocks[block] != 0) {
		printf ("duplicate index block %d\n", block);
		return -1;
	}
	index_blocks[block] = index;
	index_count++;
	return 0;
}

int
check_records (keyblk, block, size)
keyblk_t *keyblk;
off_t	block;
int	size;
{
	int	i;
	int	errors;
	off_t	record;

	for (errors = i = 0;i < keyblk->kb_cnt;i++) {
		memcpy (&record, keyblk->kb_data + i * size, sizeof record);
		if (! test_record_bit (- record)) {
			printf ("illegal record %d in block %d, slot %d\n",
				- record, block, i);
			errors++;
		}
	}
	return errors;
}

int
check_free_list (file)
file_t	*file;
{
	off_t	next = 0;
	off_t	prev = 0;
	int	size = file->f_record_size;
	char	buf[BUFSIZ];

	do {
		if (lseek (file->f_data_fd, (long) next * size, 0) == -1)
			return -1;

		if (read (file->f_data_fd, &next, sizeof next) != sizeof next)
			return -1;

again:
		if (next && test_record_bit (next)) {
			printf ("block %d is allocated and free\n", next);
			printf ("fix free list? ");
			gets (buf);
			if (buf[0] != 'y' && buf[0] != 'Y')
				continue;

			if (lseek (file->f_data_fd, (long) next * size, 0) !=
					(long) next * size)
				return -1;

			if (read (file->f_data_fd, &next, sizeof next) !=
					sizeof next)
				return -1;

			if (lseek (file->f_data_fd, (long) prev * size, 0) !=
					(long) prev * size)
				return -1;

			if (write (file->f_data_fd, &next, sizeof next) !=
					sizeof next)
				return -1;

			goto again;
		} else
			prev = next;

	} while (next != 0);

	return 0;
}

int
index_length (file, index, max)
file_t	*file;
int	index;
int	max;
{
	off_t	root;
	off_t	new_root;
	int	depth;
	int	length;
	int	i;
	int	size;
	int	errors = 0;
	keyblk_t keyblk;
	long	record_count = 0;

	size = file->f_index->i_keys[index].k_key_size + sizeof (off_t);

	for (root = index, depth = 1;root > 0;depth++) {
		printf ("            level %d:", depth);
		if (set_index_block (root, index))
			return -1;

		if (isam_read (file, root, &keyblk))
			return -1;

		memcpy (&new_root, keyblk.kb_data, sizeof root);

		if (new_root < 0)
			errors += check_records (&keyblk, root, size);

		root = new_root;
		record_count = keyblk.kb_cnt;

		for (length = 1;keyblk.kb_right;length++) {
			if (set_index_block (keyblk.kb_right, index))
				return -1;

			if (root < 0)
				errors += check_records (&keyblk,
					keyblk.kb_right, size);

			if (isam_read (file, keyblk.kb_right, &keyblk))
				return -1;

			record_count += keyblk.kb_cnt;

		}
		printf (" %d index blocks, %d entries, %4.2f%% full\n",
			length, record_count, 100.0 *
				((double) record_count) / (length * max));
	}
	return errors ? -1:0;
}

int
index_depth (file, index)
file_t	*file;
int	index;
{
	keyblk_t keyblk;
	off_t	node;
	int	depth;
	char	buf[BUFSIZ];

	for (depth = 0, node = index;node > 0;depth++) {
		if (isam_read (file, node, &keyblk)) {
			sprintf (buf, "file %s, block %d",
					file->f_file_name, node);
			db_error (buf);
			return -1;
		}
		memcpy (&node, keyblk.kb_data, sizeof node);
	}
	return depth;
}

main (argc, argv)
int	argc;
char	**argv;
{
	dbm_t	*dbm;
	file_t	*file;
	is_index_t *index;
	schema_t *schema;
	field_t	*field;
	sort_key_t *key;
	int	filenum;
	int	indexnum;
	int	i, j;
	int	max;
	int	do_dict = 0;
	long	record;
	long	errors;
	char	buf[BUFSIZ];
	struct	stat	stat;

	setbuf (stdout, 0);

	if (argc != 3) {
		if (argc == 2 && strcmp (argv[1], "-V") == 0) {
			fprintf (stderr, "dbck: version %s\n", VERSION);
			exit (0);
		}
		fprintf (stderr, "usage: dbck database file\n");
		exit (1);
	}
	if (! (dbm = db_open (argv[1], O_RDONLY))) {
		sprintf (buf, "dbck: can't open database %s", argv[1]);
		db_error (buf);
		exit (1);
	}
	if (strcmp (argv[2], "-") != 0) {
		if ((filenum = db_open_file (dbm, argv[2])) < 0) {
			sprintf (buf, "dbck: can't open file %s", argv[2]);
			db_error (buf);
			exit (1);
		}
		file = &dbm->d_files[filenum];
	} else {
		file = &dbm->d_dictionary;
		do_dict = 1;
	}
	index = file->f_index;
	schema = file->f_schema;

	if (fstat (file->f_data_fd, &stat)) {
		perror (argv[2]);
		exit (1);
	}
	max_records = stat.st_size / file->f_record_size;
	if (! (record_bits = malloc ((max_records + 7) / 8))) {
		perror ("dbck: can't allocate record bit map");
		exit (1);
	}
#ifdef BSD
	bzero (record_bits, 0, (max_records + 7) / 8);
#else
	memset (record_bits, 0, (max_records + 7) / 8);
#endif /* BSD */

	if (fstat (file->f_key_fd, &stat)) {
		perror (argv[2]); 
		exit (1);
	}
	max_index = stat.st_size / file->f_bsize;
	if (! (index_blocks = malloc (max_index))) {
		perror ("dbck: can't allocate index map\n");
		exit (1);
	}
#ifdef BSD
	bzero (index_blocks, max_index);
#else
	memset (index_blocks, 0, max_index);
#endif /* BSD */

	buf[0] = '.';
	strcpy (buf + 1, argv[2]);
	if (! do_dict) {
		if ((indexnum = db_open_index (dbm, filenum, buf)) < 0) {
			sprintf (buf, "dbck: can't open primary index .%s",
				argv[2]);
			db_error (buf);
			exit (1);
		}
	} else {
		file->f_current_index = indexnum = 1;
		isam_rewind (file);
	}
	for (errors = record_count = 0;
			(record = isam_next (file, indexnum)) != -1;
				record_count++)
		if (set_record_bit (record))
			errors++;

	if (record == -1 && db_errno != DB_EOF) {
		db_error ("dbck: premature end of file");
		exit (1);
	}
	if (errors)
		exit (1);

	if (check_free_list (file))
		exit (1);

	printf ("file %s has %ld records\n", argv[2], record_count);
	printf ("block size is %ld\n", file->f_bsize);
	printf ("there are %d indices\n", file->f_index->i_key_count);
	for (i = 1;i <= file->f_index->i_key_count;i++) {
		key = &index->i_keys[i];
		field = &schema->s_fields[key->k_field_number];
		max = (file->f_bsize - (3 * sizeof (long) + sizeof (int))) /
			(key->k_key_size + sizeof (off_t));

		printf ("    %s (%s)\n", field->f_name,
			field->f_name[0] == '.' ? "primary":"secondary");
		printf ("        key size %d, %d entries per block\n",
			key->k_key_size, max);
		printf ("        depth %d\n", index_depth (file, i));
		if (index_length (file, i, max)) {
			printf ("        index is corrupted\n");
			errors++;
		}
	}
	printf ("there are %d total index blocks used of %d total\n",
		index_count, max_index);

	if (errors) {
		printf ("one or more indexes are corrupted\n");
		exit (1);
	}
	exit (0);
}
