/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

/*
 * This file is best edited with tabstop=4.
 */

#include <sys/types.h>
#include <signal.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#ifdef CURSESX
#include <cursesX.h>
#else	/* !CURSESX */
#include <curses.h>
#endif
#ifdef BSD
#include <sgtty.h>
#define strchr index
#define memcpy(a, b, c) bcopy((b), (a), (c))
#define memcmp bcmp
#else
#include <termio.h>
#define	bzero(d,n)	memset((d), 0, (n))
#endif /* BSD */
#include <ctype.h>
#include "is_types.h"
#include "is_proto.h"
#include "patchlevel.h"
#include "isam.h"

#ifdef	MALLOC_H
#include <malloc.h>
#else
char	*malloc();
char	*realloc();
#endif

#ifndef lint
static	char	sccsid[] = "@(#)dbscreen.c	1.10	09:23:08	3/30/92";
#endif

#define	MDY_DELIMITER	'/'

#ifdef BSD
struct	sgttyb	tty_cooked;
#else
struct	termio	tty_cooked;
#endif /* BSD */
int last_cmd = 0;
int	window_offset = 0;
int	opened_window = 0;
int	screen_ok;
int	root_file;
char	*menu_bar =
"Add Begin Change Delete End Find Index Match Next Previous Quit Refresh Screen";
struct	{
	char	*name;
	int	c1;
	int	c2;
	int	match;
} matches[] = {
	"Less",		'<',	0,	IS_LESS,
	"Less-Equal",	'<',	'=',	IS_LESSEQ,
	"Equal",	'=',	0,	IS_EQUAL,
	"Greater-Equal",'>',	'=',	IS_GREATEREQ,
	"Greater",	'>',	0,	IS_GREATER,
	0,		0,	0,	0
};

long	strtol();
double	strtod();
double	floor();
double	ceil();

dbm_t	*db_open();
int	db_open_file();
long	isam_date2int();
char	*isam_int2date();

extern	int	db_errno;
extern	char	*db_errors[];
char	error_msg[BUFSIZ];

void
error (mesg)
char	*mesg;
{
	if (mesg == 0)
		mesg = db_errors[db_errno];

	move (LINES -1, 0);
	clrtoeol ();
	mvprintw (LINES - 1, 0, "%s", mesg);
	refresh ();
	getch ();
	screen_ok = 0;
}

void
set_index (dbm, filenum)
dbm_t	*dbm;
int	filenum;
{
	file_t	*file = &dbm->d_files[filenum];
	int	i;
	int	y, x;
	field_t	*field;
	char	buf[MAXFIELDNAME+1];
	char	*name;
	int	c;

	if (file->f_index == 0)
		return;

	clear ();
	move (0, 0);

	for (i = 1;i < file->f_schema->s_field_count;i++) {
		field = &file->f_schema->s_fields[i];
		if (field->f_type != T_INDEX)
			continue;

		if (field->f_name[0] == '.')
			continue;

		printw ("%.*s\n", MAXFIELDNAME, field->f_name);
	}
	while (1) {
		i = 0;
		buf[0] = '\0';
		move ((LINES-1), 0);
		clrtoeol ();
		mvaddstr ((LINES-1), 0, "Index: ");
		refresh ();

		while (i < MAXFIELDNAME && (c = getch ()) != EOF
				&& c != '\n' && c != '\r') {
			if (c == '\b') {
				if (i > 0) {
					i--;
					getyx (stdscr, y, x);
					mvaddch (y, x - 1, ' ');
					move (y, x - 1);
					buf[i] = '\0';
					refresh ();
				}
				continue;
			}
			if (! isalnum (c) && c != '_')
				continue;

			buf[i++] = c;
			buf[i] = '\0';
			addch (c);
			refresh ();
		}
		if (i == 0)
			return;

		if (name = db_index_name (dbm, filenum, buf)) {
			if ((i = db_open_index (dbm, filenum, name)) >= 0)
				return;
		}
		sprintf (error_msg, "Unable to locate an index for %s", buf);
		beep ();
		error (error_msg);
	}
}

int
get_match ()
{
	int	i;
	int	c;
	int	match;
	int	c1 = 0,
		c2 = 0;

	while (1) {
		move ((LINES-1), 0);
		clrtoeol ();
		match = -1;

		for (i = 0;matches[i].name;i++) {
			if (matches[i].c1 == c1 && matches[i].c2 == c2) {
				standout ();
				match = matches[i].match;
			}
			printw ("%s", matches[i].name);

			if (matches[i].c1 == c1 && matches[i].c2 == c2)
				standend ();

			addch (' ');
		}
		move ((LINES-1), 0);
		refresh ();

		if ((c1 || c2) && match == -1) {
			c1 = c2;
			c2 = 0;
			continue;
		}
		if ((c = getch ()) < 0) {
			match = -1;
			break;
		}
		if (c == 'g' || c == 'G')
			c = '>';
		else if (c == 'l' || c == 'L')
			c = '<';
		else if (c == 'e' || c == 'E')
			c = '=';

		if (c == '>' || c == '<' || c == '=') {
			if (match == -1) {
				c1 = c;
				c2 = 0;
			} else if (c2 != 0) {
				c1 = c2;
				c2 = c;
			} else if (c1 == 0)
				c1 = c;
			else
				c2 = c;
		} else if (c == '\r' || c == '\n') {
			break;
		} else if (c == '\b') {
			if (c2 == 0)
				c1 = 0;
			else
				c2 = 0;
		} else
			beep ();
	}
	move ((LINES-1), 0);
	clrtoeol ();
	refresh ();

	return match;
}

get_command (commands)
char	*commands;
{
	int	c;

	if ((c = getch ()) < 0)
		return -1;

	if (isupper (c))
		c = tolower (c);

	if (strchr (commands, c))
		return c;

	return -1;
}

delete_fields (dbm, screen)
dbm_t	*dbm;
screen_t *screen;
{
	int	i;

	if (screen->s_join_cnt == 0) {
		if (db_delete_record (dbm, screen->s_displays->d_fields->vf_file))
			return -1;
		else
			return 0;
	}
	if (db_delete_record (dbm, screen->s_joins->j_parent_file))
		return -1;

	for (i = 0;i < screen->s_join_cnt;i++)
		if (db_delete_record (dbm, screen->s_joins[i].j_child_file))
			return -1;

	return 0;
}

void
clear_fields (dbm, screen)
dbm_t	*dbm;
screen_t *screen;
{
	int	i;
	file_t	*file;

	if (screen->s_join_cnt == 0) {
		file = &dbm->d_files[screen->s_displays->d_fields->vf_file];
		bzero (file->f_data_buffer, file->f_record_size);
	}
	file = &dbm->d_files[screen->s_joins->j_parent_file];
	bzero (file->f_data_buffer, file->f_record_size);

	for (i = 0;i < screen->s_join_cnt;i++) {
		file = &dbm->d_files[screen->s_joins[i].j_child_file];
		bzero (file->f_data_buffer, file->f_record_size);
	}
}

update_fields (dbm, screen)
dbm_t	*dbm;
screen_t *screen;
{
	int	i;
	file_t	*file;

	if (screen->s_join_cnt == 0) {
		file = &dbm->d_files[screen->s_displays->d_fields->vf_file];
		if (db_update_record (dbm,
				screen->s_displays->d_fields->vf_file,
				file->f_data_buffer))
			return -1;
		else
			return 0;
	}
	file = &dbm->d_files[screen->s_joins->j_parent_file];
	if (db_update_record (dbm, screen->s_joins->j_parent_file,
			file->f_data_buffer))
		return -1;

	for (i = 0;i < screen->s_join_cnt;i++) {
		file = &dbm->d_files[screen->s_joins[i].j_child_file];
		if (db_update_record (dbm, screen->s_joins[i].j_child_file,
				file->f_data_buffer))
			return -1;
	}
	return 0;
}

/*
 * enter_fields - input data into a named field
 *
 * enter_fields() allows the user to type data into a field either for
 * adding a new record or updating an existing record.
 */

enter_fields (dbm, display)
dbm_t	*dbm;
display_t *display;
{
	file_t	*file;
	field_t	*field;
	vfield_t *vfield;
	char	*cp;
	int	i;

	/*
	 * Update every field in the current display
	 */

	for (i = 0;i < display->d_field_cnt;i++) {

		/*
		 * Pick the next field.  If it read-only, skip it and
		 * go again.
		 */

		vfield = &display->d_fields[i];
		if (vfield->vf_readonly)
			continue;

		/*
		 * Determine the file the field is a part of and which
		 * field in that file is being altered.  Index fields
		 * may not be updated - return an error if one gets into
		 * the list somehow ...
		 */

		file = &dbm->d_files[vfield->vf_file];
		field = &file->f_schema->s_fields[vfield->vf_field];

		if (field->f_type == T_INDEX)
			return -1;

		/*
		 * Locate the data in the file buffer and call the entry
		 * routine to update it.
		 */

		cp = file->f_data_buffer + field->f_offset;
		if (enter_field (field, cp, vfield->vf_y, vfield->vf_x))
			return -1;
	}
	return 0;
}

display_fields (dbm, display)
dbm_t	*dbm;
display_t *display;
{
	file_t	*file;
	field_t	*field;
	vfield_t *vfield;
	int	i;
	int	in_bracket;
	int	y;
	char	*cp;

	if (display->d_width > COLS || display->d_height > LINES)
		return -1;

	for (y = 0, cp = display->d_background;*cp;cp++, y++) {
		move (y, 0);
		clrtoeol ();
		move (y, 0);

		in_bracket = 0;
		while (*cp) {
			if (*cp == ']')
				in_bracket = 0;

			if (in_bracket) {
				if (! isspace (*cp))
					addch (' ');
				else
					addch (*cp);
			} else if (*cp == '[') {
				in_bracket = 1;
				addch ('[');
			} else
				addch (*cp);

			cp++;
		}
	}
	for (i = 0;i < display->d_field_cnt;i++) {
		vfield = &display->d_fields[i];
		file = &dbm->d_files[vfield->vf_file];
		field = &file->f_schema->s_fields[vfield->vf_field];

		if (display_field (file, field, vfield->vf_y, vfield->vf_x))
			return -1;
	}
	refresh ();
	return 0;
}

read_next (dbm, screen)
dbm_t	*dbm;
screen_t *screen;
{
	int	i;

	if (screen->s_join_cnt == 0) {
		if (db_read_record (dbm, root_file, (void *) 0, IS_NEXT))
			goto failure;
		else
			return 0;
	}
	for (i = screen->s_join_cnt - 1;i >= 0;i--) {
		if (! read_join_next (dbm, screen->s_joins + i, 1))
			break;
	}
	if (i < 0 && db_read_record (dbm, root_file, (void *) 0, IS_NEXT))
		goto failure;

	i++;
	for (;i < screen->s_join_cnt;i++)
		if (read_join (dbm, screen->s_joins + i))
			goto failure;

	return 0;

failure:
	clear_fields (dbm, screen);
	(void) read_fields (dbm, screen);
	return -1;
}

read_previous (dbm, screen)
dbm_t	*dbm;
screen_t *screen;
{
	int	i;

	if (screen->s_join_cnt == 0) {
		if (db_read_record (dbm, root_file, (void *) 0, IS_PREV))
			goto failure;
		else
			return 0;
	}
	for (i = screen->s_join_cnt - 1;i >= 0;i--) {
		if (! read_join_next (dbm, screen->s_joins + i, -1))
			break;
	}
	if (i < 0 && db_read_record (dbm, root_file, (void *) 0, IS_PREV))
		goto failure;

	i++;
	for (;i < screen->s_join_cnt;i++)
		if (read_join (dbm, screen->s_joins + i))
			goto failure;

	return 0;

failure:
	clear_fields (dbm, screen);
	(void) read_fields (dbm, screen);
	return -1;
}

read_file (dbm, filenum)
dbm_t	*dbm;
int	filenum;
{
	return db_read_record (dbm, filenum, (void *) 0, IS_CURR);
}

read_join (dbm, join)
dbm_t	*dbm;
join_t	*join;
{
	char	buf[1024];
	char	*cp;
	int	i;
	file_t *file;
	field_t	*field;
	sort_key_t *key;

	file = &dbm->d_files[join->j_parent_file];
	field = &file->f_schema->s_fields[join->j_parent_field];

	if (field->f_type == T_INDEX) {
		key = &file->f_index->i_keys[field->f_offset];

		for (cp = buf, i = 0;i < key->k_field_count;i++) {
			if (! key->k_unique && i == key->k_field_count - 1)
				break;

			field = &file->f_schema->s_fields[key->k_fields[i]];
			memcpy (cp, file->f_data_buffer +
					field->f_offset, field->f_size);
			cp += field->f_size;
		}
		cp = buf;
	} else
		cp = file->f_data_buffer + field->f_offset;

	if (db_read_record (dbm, join->j_child_file, (void *) cp, IS_EQUAL))
		return -1;

	return 0;
}

read_join_next (dbm, join, dir)
dbm_t	*dbm;
join_t	*join;
int	dir;
{
	char	buf1[1024];
	char	buf2[1024];
	int	key_size;
	char	*key1;
	char	*key2;
	char	*cp;
	int	i;
	file_t *file;
	field_t	*field;
	sort_key_t *key;

	file = &dbm->d_files[join->j_parent_file];
	field = &file->f_schema->s_fields[join->j_parent_field];

	if (field->f_type == T_INDEX) {
		key = &file->f_index->i_keys[field->f_offset];

		for (cp = buf1, i = 0;i < key->k_field_count;i++) {
			if (! key->k_unique && i == key->k_field_count - 1)
				break;

			field = &file->f_schema->s_fields[key->k_fields[i]];
			memcpy (cp, file->f_data_buffer +
					field->f_offset, field->f_size);
			cp += field->f_size;
		}
		key_size = cp - buf1;
		key1 = buf1;
	} else {
		key1 = file->f_data_buffer + field->f_offset;
		key_size = field->f_size;
	}
	if (db_read_record (dbm, join->j_child_file, (void *) 0,
			dir > 0 ? IS_NEXT:IS_PREV))
		return -1;

	file = &dbm->d_files[join->j_child_file];
	key = &file->f_index->i_keys[join->j_child_field];

	for (cp = buf2, i = 0;i < key->k_field_count;i++) {
		if (! key->k_unique && i == key->k_field_count - 1)
			break;

		field = &file->f_schema->s_fields[key->k_fields[i]];
		memcpy (cp, file->f_data_buffer +
				field->f_offset, field->f_size);
		cp += field->f_size;
	}
	key2 = buf2;

	if (memcmp (key1, key2, key_size))
		return 1;
	else
		return 0;
}

reread_fields (dbm, screen)
dbm_t	*dbm;
screen_t *screen;
{
	int	i;

	if (screen->s_join_cnt == 0) {
		if (read_file (dbm, screen->s_displays->d_fields->vf_file))
			return -1;
		else
			return 0;
	}
	if (read_file (dbm, screen->s_joins->j_parent_file))
		return -1;

	for (i = 0;i < screen->s_join_cnt;i++)
		if (read_file (dbm, screen->s_joins[i].j_child_file))
			return -1;

	return 0;
}

read_fields (dbm, screen)
dbm_t	*dbm;
screen_t *screen;
{
	int	i;

	clear_fields (dbm, screen);
	if (screen->s_join_cnt == 0) {
		if (read_file (dbm, screen->s_displays->d_fields->vf_file))
			return -1;
		else
			return 0;
	}
	if (read_file (dbm, screen->s_joins->j_parent_file))
		return -1;

	for (i = 0;i < screen->s_join_cnt;i++)
		if (read_join (dbm, &screen->s_joins[i]))
			return -1;

	return 0;
}

write_file (dbm, filenum)
dbm_t	*dbm;
int	filenum;
{
	return db_write_record (dbm, filenum, dbm->d_files[filenum].f_data_buffer);
}

write_fields (dbm, screen)
dbm_t	*dbm;
screen_t *screen;
{
	int	i;

	if (screen->s_join_cnt == 0) {
		if (write_file (dbm, screen->s_displays->d_fields->vf_file))
			return -1;
		else
			return 0;
	}
	if (write_file (dbm, screen->s_joins->j_parent_file))
		return -1;

	for (i = 0;i < screen->s_join_cnt;i++)
		if (write_file (dbm, screen->s_joins[i].j_child_file))
			return -1;

	return 0;
}

draw_menu (menu, command)
char	*menu;
int	command;
{
	char 	*cp;
	int	stand_out = 0;

	if (command != 0 && islower (command))
		command = toupper (command);

	move ((LINES-1), 0);
	clrtoeol ();

	for (cp = menu;*cp;cp++) {
		if (command && command == *cp) {
			standout ();
			stand_out = 1;
		}
		if (stand_out && *cp == ' ') {
			standend ();
			stand_out = 0;
		}
		addch (*cp);
	}
	if (stand_out)
		standend ();
}

draw_field (field, line)
field_t	*field;
int	line;
{
	int	j;
	int	x, y;

	mvprintw (line, 0, "%.*s", MAXFIELDNAME, field->f_name);
	addch (' ');
	addch ('[');
	switch (field->f_type) {
		case T_CHAR:
			for (j = 0;j < field->f_size;j++) {
				getyx (stdscr, y, x);
				if (x >= (COLS-2)) {
					addch ('>');
					break;
				}
				addch (' ');
			}
			break;
		case T_SHORT:
			for (j = 0;j < 6;j++) addch (' ');
			break;
		case T_INT:
			if (sizeof (int) == sizeof (short))
				for (j = 0;j < 6;j++) addch (' ');
			else
				for (j = 0;j < 10;j++) addch (' ');
			break;
		case T_LONG:
		case T_SERIAL:
			for (j = 0;j < 10;j++) addch (' ');
			break;
		case T_DATE:
			for (j = 0;j < 8;j++) addch (' ');
			break;
		case T_FLOAT:
			for (j = 0;j < 12;j++) addch (' ');
			break;
		case T_DOUBLE:
			for (j = 0;j < 22;j++) addch (' ');
			break;
		case T_MONEY:
			for (j = 0;j < 17;j++) addch (' ');
			break;
	}
	addch (']');
}

display_field (file, field, row, column)
file_t	*file;
field_t	*field;
int	row;
int	column;
{
	short	s;
	int	i;
	long	l;
	float	f;
	double	d;
	char	*cp;

	switch (field->f_type) {
		case T_SHORT:
			memcpy (&s, file->f_data_buffer + field->f_offset,
				sizeof s);
			mvprintw (row, column, "%6d", s);
			break;
		case T_INT:
			memcpy (&i, file->f_data_buffer + field->f_offset,
				sizeof i);
			if (sizeof (int) == sizeof (short))
				mvprintw (row, column, "%6d", i);
			else
				mvprintw (row, column, "%10d", i);
			break;
		case T_LONG:
		case T_SERIAL:
			memcpy (&l, file->f_data_buffer + field->f_offset,
				sizeof l);
			mvprintw (row, column, "%10d", l);
			break;
		case T_DATE:
			memcpy (&l, file->f_data_buffer + field->f_offset,
				sizeof l);
			mvprintw (row, column, "%.10s",
				isam_int2date (l, MDY_DELIMITER));
			break;
		case T_FLOAT:
			memcpy (&f, file->f_data_buffer + field->f_offset,
				sizeof f);
			mvprintw (row, column, "%12g", f);
			break;
		case T_DOUBLE:
			memcpy (&d, file->f_data_buffer + field->f_offset,
				sizeof d);
			mvprintw (row, column, "%22g", d);
			break;
		case T_MONEY:
			memcpy (&d, file->f_data_buffer + field->f_offset,
				sizeof d);
			mvprintw (row, column, "%17.2f", d);
			break;
		case T_CHAR:
			cp = file->f_data_buffer + field->f_offset;
			move (row, column);
			for (i = 0;cp[i] != '\0' && i < field->f_size &&
					column + i < (COLS-2);i++) {
				if (isprint (cp[i]))
					addch (cp[i]);
				else
					addch ('?');
			}
			for (;i < field->f_size && column + i < (COLS-2);i++)
				addch (' ');
			break;
		case T_INDEX:
			return -1;
	}
	return 0;
}

enter_field (field, data, row, column)
field_t	*field;
char	*data;
int	row;
int	column;
{
	double	d;
	float	f;
	short	s;
	long	l;
	char	buf[BUFSIZ];
	char	*ep;
	int	i;
	int	x, y;
	int	c;
	int	len;

	switch (field->f_type) {
		case T_CHAR:	len = field->f_size; break;
		case T_SHORT:	len = 6; break;
		case T_INT:		if (sizeof (int) == sizeof (short))
							len = 6;
						else
							len = 10;
						break;
		case T_SERIAL:	return 0;
		case T_DATE:
		case T_LONG:	len = 10; break;
		case T_FLOAT:	len = 12; break;
		case T_MONEY:	len = 17; break;
		case T_DOUBLE:	len = 22; break;
		default:		return -1;
	}
	move (row, column);
	for (i = 0;i < len && i + column < (COLS-2);i++)
		addch (' ');

	move (row, column);
	refresh ();

	i = 0;
	while ((c = getch ()) != EOF && c != '\n' &&
			c != '\r' && i < len && i + column < (COLS-1)) {
		if (c == '\b') {
			if (i == 0)
				continue;

			getyx (stdscr, y, x);
			mvaddch (y, x - 1, ' ');
			move (y, x - 1);
			buf[--i] = '\0';
			refresh ();
			continue;
		}
		addch (c);
		buf[i++] = c;
		buf[i] = 0;
		refresh ();
	}
	if (i == 0)
		return 0;

	switch (field->f_type) {
		case T_SHORT:
do_short:
			s = (short) strtol (buf, &ep, 10);
			if (*ep != '\0')
				return -1;

			memcpy (data, &s, sizeof s);
			mvprintw (row, column, "%6d", s);
			break;
		case T_INT:
			if (sizeof (int) == sizeof (short))
				goto do_short;
			else
				goto do_long;
		case T_LONG:
		case T_SERIAL:
do_long:
			l = strtol (buf, &ep, 10);
			if (*ep != '\0')
				return -1;

			memcpy (data, &l, sizeof l);
			mvprintw (row, column, "%10d", l);
			break;
		case T_DATE:
			if ((l = isam_date2int (buf, MDY_DELIMITER)) == 0)
				return 0;

			memcpy (data, (char *) &l, sizeof l);
			mvprintw (row, column, "%.10s", isam_int2date (l, MDY_DELIMITER));
			break;
		case T_FLOAT:
			f = strtod (buf, &ep);
			if (*ep != '\0')
				return -1;

			memcpy (data, (char *) &f, sizeof f);
			mvprintw (row, column, "%12.6g", f);
			break;
		case T_DOUBLE:
			d = strtod (buf, &ep);
			if (*ep != '\0')
				return -1;

			memcpy (data, (char *) &d, sizeof d);
			mvprintw (row, column, "%22.11g", d);
			break;
		case T_MONEY:
			d = strtod (buf, &ep);
			if (*ep != '\0')
				return -1;

			if (d > 0)
				d = floor (d * 100.0 + 0.5);
			else
				d = ceil (d * 100.0 - 0.5);

			memcpy (data, (char *) &d, sizeof d);
			mvprintw (row, column, "%17.2f", d / 100.0);
			break;
		case T_CHAR:
			bzero (data, field->f_size);
			memcpy (data, buf, i);
			mvprintw (row, column, "%.*s", field->f_size, data);
			break;
		case T_INDEX:
			return -1;
	}
	return 0;
}

enter_index (file, buf)
file_t	*file;
char	*buf;
{
	char	*cp;
	int	indexnum = file->f_current_index;
	int	i;
	int	max;
	int	column;
	field_t	*field;
	sort_key_t *key;

	if (file->f_index == 0 || indexnum <= 1 ||
			indexnum > file->f_index->i_key_count) {
		error ("No current index");
		return -1;
	}
	key = file->f_index->i_keys + indexnum;

	if (key->k_key_size > BUFSIZ)
		return -1;

	max = key->k_field_count;
	if (! key->k_unique)
		max--;

	bzero (buf, key->k_key_size);
	for (i = 0;i < (LINES-1);i++) {
		move (i, 0);
		clrtoeol ();
	}
	for (i = 0;i < max;i++) {
		field = &file->f_schema->s_fields[key->k_fields[i]];
		draw_field (field, i);
	}
	refresh ();
	for (cp = buf, i = 0;i < max;i++) {
		field = &file->f_schema->s_fields[key->k_fields[i]];
		if ((column = strlen (field->f_name)) >= MAXFIELDNAME)
			column = MAXFIELDNAME;

		column += 2;
		if (enter_field (field, cp, i, column))
			return -1;

		cp += field->f_size;
	}
	return 0;
}

do_menu (dbm, screen)
dbm_t	*dbm;
screen_t *screen;
{
	char	buf[BUFSIZ];
	int	command;
	int	i;
	int	match = IS_EQUAL;
	int	record_ok = 0;
	display_t *display;
	file_t	*file;

	display = screen->s_displays;

	while (1) {
		if (! record_ok)
			clear_fields (dbm, screen);

		if (! screen_ok) {
			clear ();

			draw_menu (menu_bar, 0);
			(void) display_fields (dbm, display);

			screen_ok = 1;
		} else {
			draw_menu (menu_bar, 0);
		}
		move ((LINES-1), 0);
		refresh ();

		if ((command = get_command ("abcdefimnpqrs")) < 0) {
			beep ();
			continue;
		} else {
			draw_menu (menu_bar, command);
			move ((LINES-1), 0);
			refresh ();
		}
		switch (command) {
			case 'a':
				if (dbm->d_mode == O_RDONLY) {
					beep ();
					error ("Read-only database");
					break;
				}
				clear_fields (dbm, screen);

				if (! display_fields (dbm, display) &&
						! enter_fields (dbm, display)) {
						if (write_fields (dbm, screen)) {
							beep ();
							error ("Unable to write record");
						}
						record_ok = 0;
				} else if (reread_fields (dbm, screen)) {
					beep ();
					error ("Unable to re-read current record");
					record_ok = 0;
				} else
					record_ok = 1;

				screen_ok = 0;
				break;
			case 'b':
				if (db_rewind (dbm, root_file) ||
						read_next (dbm, screen)) {
					beep ();
					error ("Unable to read from BOF");
					record_ok = 0;
				}
				screen_ok = 0;
				break;
			case 'c':
				if (dbm->d_mode == O_RDONLY) {
					beep ();
					error ("Read-only database");
					break;
				}
				if (enter_fields (dbm, display) ||
						update_fields (dbm, screen)) {
					beep ();
					error ("Unable to update record");
					record_ok = 0;
				}
				break;
			case 'd':
				if (dbm->d_mode == O_RDONLY) {
					beep ();
					error ("Read-only database");
				} else if (delete_fields (dbm, screen)) {
					beep ();
					error ("Unable to delete record");
					record_ok = 0;
				} else
					record_ok = 0;

				screen_ok = 0;
				break;
			case 'e':
				if (db_rewind (dbm, root_file) ||
						read_previous (dbm, screen)) {
					beep ();
					error ("Unable to read from EOF");
					record_ok = 0;
				}
				screen_ok = 0;
				break;
			case 'f':
				file = dbm->d_files + root_file;
				record_ok = 0;

				if (file->f_current_index <= 1) {
					error ("No index defined.  Use the index command");
					set_index (dbm, root_file);

					if (file->f_current_index <= 1)
						break;
				}
				if (file->f_current_index <= 0 ||
					enter_index (file, buf) ||
					db_read_record (dbm, root_file, (void *) buf, match) ||
						read_fields (dbm, screen)) {
					beep ();
					error ("Unable to find matching record");
				} else
					record_ok = 1;

				screen_ok = 0;
				break;
			case 'i':
				set_index (dbm, root_file);
				screen_ok = 0;
				break;
			case 'm':
				if ((i = get_match ()) != -1)
					match = i;

				break;
			case 'n':
				if (read_next (dbm, screen)) {
					beep ();
					if (db_errno != DB_EOF) {
						error ("Unable to read next record");
						record_ok = 0;
					}
				} else
					record_ok = 1;

				screen_ok = 0;
				break;
			case 'p':
				if (read_previous (dbm, screen)) {
					beep ();
					if (db_errno != DB_BOF) {
						error ("Unable to read previous record");
						record_ok = 0;
					}
				} else
					record_ok = 1;

				screen_ok = 0;
				break;
			case 'q':
				return;
			case 'r':
				if (! record_ok)
					break;

				if (reread_fields (dbm, screen)) {
					beep ();
					error ("Unable to read current record");
					record_ok = 0;
				} else
					record_ok = 1;

				screen_ok = 0;
				break;
			case 's':
				display++;
				if (display >= screen->s_displays +
						screen->s_display_cnt)
					display = screen->s_displays;

				screen_ok = 0;
				break;
		}
	}
}

/*
 * usage - display command line syntax and exit
 */

usage ()
{
	(void) fprintf (stderr, "usage: dbscreen { -V | [ -r ] screen_file }\n");
	exit (1);
}

#ifdef	VOIDSIG
void
#else
int
#endif
catch (sig)
int	sig;
{
#ifndef	NO_IOCTL
#ifdef BSD
	ioctl (0, TIOCSETP, &tty_cooked);
#else
	ioctl (0, TCSETA, &tty_cooked);
#endif /* BSD */
#endif /* !NO_IOCTL */
	if (opened_window) {
		endwin ();
		move ((LINES-1), 0);
		refresh ();
	}
	exit (sig);
}

display_size (display)
char	*display;
{
	int	i;
	int	j;

	for (i = 0;*display;j = strlen (display) + 1, i += j, display += j)
		;

	return i + 1;
}

swap_joins (join_a, join_b)
join_t	*join_a;
join_t	*join_b;
{
	join_t	join_tmp;

	join_tmp = *join_a;
	*join_a = *join_b;
	*join_b = join_tmp;
}

screen_t *
open_screen (dbm, bscreen)
dbm_t	*dbm;
bscreen_t *bscreen;
{
	screen_t *screen;
	display_t *display;
	bdisplay_t *bdisplay;
	vfield_t *field;
	bfield_t *bfield;
	join_t	*join;
	int	i, j;
	int	len;

	if (! (screen = (screen_t *) malloc (sizeof *screen)))
		return 0;

	bzero (screen, sizeof *screen);

	screen->s_display_cnt = bscreen->bs_screen_cnt;
	screen->s_join_cnt = 0;
	screen->s_joins = 0;

	if (! (screen->s_displays = (display_t *)
			malloc (sizeof (display_t) * screen->s_display_cnt)))
		return 0;

	bzero (screen->s_displays, sizeof (display_t) * screen->s_display_cnt);

	for (i = 0, bdisplay = bscreen->bs_screens;i < screen->s_display_cnt;
			i++, bdisplay = bd_next (bdisplay)) {

		display = &screen->s_displays[i];

		display->d_height = bdisplay->bd_height;
		display->d_width = bdisplay->bd_width;
		display->d_field_cnt = 0;

		len = display_size (bdisplay->bd_background);
		if (! (display->d_background = (char *) malloc (len)))
			return 0;
		memcpy (display->d_background, bdisplay->bd_background, len);

		for (j = 0, bfield = bdisplay->bd_fields;
				j < bdisplay->bd_field_cnt;
				j++, bfield = bf_next (bfield)) {

			/*
			 * See if this is a dummy field that describes how two
			 * files are related.  They don't appear on the screen,
			 * so we just get the file and field info and stuff it
			 * in the join array
			 */

			if (bfield->bf_join) {

				/*
				 * Allocate storage for the next join
				 */

				if (screen->s_joins) {
					screen->s_joins = (join_t *) realloc (screen->s_joins,
								(screen->s_join_cnt + 1) * sizeof (join_t));
				} else {
					screen->s_joins = (join_t *) malloc (sizeof (join_t));
					screen->s_join_cnt = 0;
				}
				if (screen->s_joins == 0)
					return 0;

				join = &screen->s_joins[screen->s_join_cnt++];

				/*
				 * Open the file and field for the child record.  This is
				 * the file that will be looked up in to get the next
				 * chunk of data.
				 */

				if ((join->j_child_file = db_open_file (dbm,
						bfield->bf_file)) < 0)
					return 0;

				if ((join->j_child_field = db_open_index (dbm,
						join->j_child_file, bfield->bf_field)) < 0)
					return 0;

				/*
				 * Open the file and field for the parent record.  This
				 * is the file that will be used to locate the child
				 * record.
				 */

				if ((join->j_parent_file = db_open_file (dbm,
						bfield->bf_join_file)) < 0)
					return 0;

				if ((join->j_parent_field = db_open_field (dbm,
						join->j_parent_file, bfield->bf_join_field)) < 0)
					return 0;

				/*
				 * Since this is just a join, skip to the next entry
				 * and process it (if there is one)
				 */

				continue;
			}
			display->d_field_cnt++;

			/*
			 * Allocate storage for the next screen field
			 */

			if (! display->d_fields) {
				if (! (display->d_fields = (vfield_t *) malloc
						(sizeof (vfield_t) * display->d_field_cnt)))
					return 0;
			} else {
				if (! (display->d_fields = (vfield_t *) realloc
						(display->d_fields, sizeof (vfield_t) *
						display->d_field_cnt)))
					return 0;
			}
			field = &display->d_fields[display->d_field_cnt - 1];

			/*
			 * Initialize the fields parts.  The coordinates and
			 * read/write status come from the screen file and the
			 * file and field numbers are gotton from the dictionary.
			 */

			field->vf_x = bfield->bf_x;
			field->vf_y = bfield->bf_y;
			field->vf_readonly = bfield->bf_readonly;

			if ((field->vf_file = db_open_file (dbm,
					bfield->bf_file)) < 0)
				return 0;

			if ((field->vf_field = db_open_field (dbm,
					field->vf_file, bfield->bf_field)) < 0)
				return 0;
		}
	}

	/*
	 * Re-order any joins so that the parent files are before the
	 * child files in all the joins.  This way the first file in
	 * the join list can be read, and that data can be used to read
	 * the next file, and so on.
	 */

	if (screen->s_join_cnt) {
		join = screen->s_joins;

		for (i = 0;i < screen->s_join_cnt;i++)
			for (j = i + 1;j < screen->s_join_cnt;j++)
				if (join[i].j_parent_file == join[j].j_child_file)
					swap_joins (&join[i], &join[j]);
	}
	return screen;
}

#ifdef BSD
beep()
{
    return write(2, "\007", 1);
}
#endif /* BSD */

/* 
 * dbscreen - add, change, or delete records from one or more database files
 */

main (argc, argv)
int	argc;
char	**argv;
{
	dbm_t	*dbm;
	FILE	*scr;
	int	mode = O_RDWR;
	bscreen_t *bscreen;
	screen_t *screen;
	char	buf[BUFSIZ];
	char	file[BUFSIZ];
	char	path[BUFSIZ];
	char	pri_index[MAXFIELDNAME+1];
	char	*cindex = (char *) 0;
	char	*cp, *cp2;
	extern	int	optind;
	extern	char	*optarg;
	int	flag;
	int	i;

	/*
	 * Get the flags from the command line
	 *
	 *	-V - display version number
	 *	-i - set default index
	 *	-r - set read-only mode
	 */

	while ((flag = getopt (argc, argv, "ri:V")) != EOF) {
		switch (flag) {
			case 'V':
				fprintf (stderr, "dbscreen: version %s\n", VERSION);
				exit (0);
			case 'i':
				cindex = optarg;
				break;
			case 'r':
				mode = O_RDONLY;
				break;
			default:
				usage ();
		}
	}
	if (optind + 1 != argc)
		usage ();

	/*
	 * Open the user's screen file and read in the parts.  The file
	 * must end in ".scr" and will be located using DBPATH, as for
	 * isam_file_open.
	 */

	if (! (cp = strrchr (argv[optind], '.')) || strcmp (cp, ".scr") != 0)
		strcat (strcpy (file, argv[optind]), ".scr");
	else
		strcpy (file, argv[optind]);

	if (file[0] != '/') {
		if (! (cp = getenv ("DBPATH")))
			cp = ".";

		do {
			if (cp2 = strchr (cp, ':')) {
				strncpy (path, cp, cp2 - cp);
				path[cp2 - cp] = '\0';
				cp2++;
			} else
				strcpy (path, cp);

			strcat (path, "/");
			strcat (path, file);
		} while (access (path, 0) && (cp = cp2));
	} else
		strcpy (path, file);

	/*
	 * The absolute file name has been figure out, complete with
	 * the trailing ".scr" extension.
	 */

	if ((scr = fopen (path, "r")) == 0) {
		(void) sprintf (error_msg, "dbscreen: can't open screen %s",
			file);
		perror (error_msg);
		exit (1);
	}

	/*
	 * Allocate space for, and read in the file header.
	 */

	if (! (bscreen = (bscreen_t *) malloc (sizeof *bscreen))) {
		perror ("dbscreen: out of memory");
		exit (1);
	}
	if (fread (&bscreen->bs_size, sizeof bscreen->bs_size, 1, scr) != 1) {
		perror ("dbscreen: cannot read file header");
		exit (1);
	}

	/*
	 * Take the object size from the header, and re-allocate enough
	 * space to store the entire object.
	 */

	if (! (bscreen = (bscreen_t *) realloc (bscreen, bscreen->bs_size))) {
		perror ("dbscreen: out of memory");
		exit (1);
	}
	rewind (scr);
	if (fread (bscreen, 1, bscreen->bs_size, scr) != bscreen->bs_size) {
		perror ("dbscreen: cannot read screen file");
		exit (1);
	}

	/*
	 * Open the database which was named in the screen file
	 */

	if (! (dbm = db_open (bscreen->bs_database, mode))) {
		(void) sprintf (error_msg, "dbscreen: can't open database %s",
			bscreen->bs_database);
		db_error (error_msg);
		exit (1);
	}

	/* 
	 * Get all of the parts of the screen and open all the appropriate
	 * files, etc.
	 */

	if (! (screen = open_screen (dbm, bscreen)))
		exit (1);

	/*
	 * Open the default index.  The default index is in the file
	 * which is the first parent of the first join, or if there
	 * are no joins, there is only one file, which is then used.
	 */

	if (screen->s_join_cnt == 0)
		root_file = screen->s_displays->d_fields->vf_file;
	else
		root_file = screen->s_joins->j_parent_file;

	if (! cindex) {
		sprintf (pri_index, ".%.*s", MAXFIELDNAME,
				dbm->d_files[root_file].f_file_name);
		cindex = pri_index;
	}
	if (db_open_index (dbm, root_file, cindex) < 0) {
		sprintf (buf, "dbscreen: can't open index %s", cindex);
		db_error (buf);
		exit (1);
	}

	/* 
	 * Save the current TTY modes so that they can be restored later
	 * on.  The signal catching routine needs to know that the modes
	 * have been saved so it can restore them if it has to.
	 */

#ifdef BSD
	ioctl (0, TIOCGETP, &tty_cooked);
#else
	ioctl (0, TCGETA, &tty_cooked);
#endif /* BSD */

	/*
	 * Initialize curses and set the operating modes
	 */

	initscr ();
	opened_window = 1;
	nonl ();
#ifdef BSD
	crmode ();
#else	
	cbreak ();
#endif /* BSD */
	noecho ();

	/* 
	 * Set up the signal catchers
	 */

#ifdef	SIGINT
	(void) signal (SIGINT, catch);
#endif
#ifdef	SIGQUIT
	(void) signal (SIGQUIT, catch);
#endif

	/*
	 * Display the menu and start accepting user commands
	 */

	do_menu (dbm, screen);

	/*
	 * Move to the lower left-hand corner so the user's prompt
	 * won't appear in the middle of the old screen.
	 */

	move ((LINES-1), 0);
	refresh ();
	endwin ();

	/* 
	 * Restore the original modes, give the user a fresh new line
	 * and exit.
	 */

#ifdef BSD
	ioctl (0, TIOCSETP, &tty_cooked);
#else
	ioctl (0, TCSETA, &tty_cooked);
#endif /* BSD */

	putchar ('\n');
	exit (0);
	/*NOTREACHED*/
}
