/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <stdio.h>
#ifdef	BSD
#include <strings.h>
#define	strchr	index
#define	strrchr	rindex
#else
#include <string.h>
#ifndef	SYSIII
#include <memory.h>
#endif
#endif
#ifdef	STDLIB_H
#include <stdlib.h>
#endif
#include "is_types.h"
#include "is_proto.h"

#ifdef BSD
#define memcpy(a, b, c) bcopy((b), (a), (c))
#endif /* BSD */

#ifndef lint
static	char	sccsid[] = "@(#)is_del.c	1.6	10:32:19	2/26/92";
#endif

#ifndef	STDLIB_H
extern	char	*getenv();
#endif

extern	int	db_errno;

__isam_delete (file, key, block, offset, node)
file_t	*file;
sort_key_t *key;
off_t	block;
int	offset;
keyblk_t *node;
{
	keyblk_t keyblk;
	off_t	parent, tmp;
	int	i;
	int	size = key->k_key_size + sizeof (off_t);

top:
	if (offset > 0) {

		/*
		 * Node isn't first in block, so it can't be the
		 * last one remaining.  I need to fix the index
		 * stuff because the key might be the last key in
		 * this block, in which case the current key
		 * becomes the first key in the block to the right.
		 */

		__isam_key_delete (node, key, offset, 1);

		if (block == file->f_key_block) {
			if (offset >= node->kb_cnt) {
				file->f_key_block = node->kb_right;
				file->f_key_slot = 0;
			}
		}
		return isam_write (file, block, node);
	}
	if (node->kb_cnt > 0) {

		/*
		 * Node is the first one in the block, but not the
		 * last.  I don't have to change the index stuff
		 * because the key that gets moved over the deleted
		 * key will become the next key if things just get
		 * left alone.
		 */

		__isam_key_delete (node, key, 0, 1);
		if (isam_write (file, block, node))
			return -1;

		if (node->kb_prev > 0) {
			if (__isam_fix (file, key, node, block, 0) == -1)
				return -1;
			else
				return 0;
		} else
			return 0;
	}

	/*
	 * This is the first and only key in this block.  The index
	 * stuff needs to be fixed up to point to the first key in
	 * the right sibbling.  The left and right sibblings need to
	 * be pointed at each other, and this node needs to be freed.
	 * Then we recursively delete the reference to this node
	 * from the parent node.
	 */

	if (block == file->f_key_block) {
		file->f_key_block = node->kb_right;
		file->f_key_slot = 0;
	}
	if (node->kb_left > 0) {
		if (isam_read (file, node->kb_left, &keyblk))
			return -1;

		keyblk.kb_right = node->kb_right;
		if (isam_write (file, node->kb_left, &keyblk))
			return -1;
	}
	if (node->kb_right > 0) {
		if (isam_read (file, node->kb_right, &keyblk))
			return -1;

		keyblk.kb_left = node->kb_left;
		if (isam_write (file, node->kb_right, &keyblk))
			return -1;
	}
	parent = node->kb_prev;

	if (parent == 0) {

		/*
		 * You can't free the root node, but you can make
		 * it empty.
		 */

		node->kb_cnt = 0;
		node->kb_left = 0;
		node->kb_right = 0;
		return isam_write (file, block, node);
	}
	if (isam_free (file, offset))
		return -1;

	return __isam_delete_block (file, key, parent, block);
}

/*
 * __isam_delete_block - delete key associate with `block'
 *
 * __isam_delete_block() scans the current index block, looking for the
 * key associated with the `block' parameter.  That key is then deleted
 * from the index.
 */

int
__isam_delete_block (file, key, parent, block)
file_t	*file;
sort_key_t *key;
off_t	parent;
off_t	block;
{
	keyblk_t node;
	off_t	tmp;
	int	i;
	int	size = key->k_key_size + sizeof (off_t);

	if (isam_read (file, parent, &node))
		return -1;

	for (i = 0;i < node.kb_cnt;i++) {
		memcpy (&tmp, node.kb_data + i * size, sizeof tmp);
		if (tmp == block)
			break;
	}
	if (i == node.kb_cnt)
		return -1;

	return __isam_delete (file, key, parent, i, &node);
}

isam_delete (file, index, data, record)
file_t	*file;
int	index;
VOIDP	*data;
off_t	record;
{
	sort_key_t *key = &file->f_index->i_keys[index];
	int	rc;
	long	ticket;
	off_t	record2;
	int	was_locked = file->f_locked;
	off_t	old_block;

	if (! was_locked) {
		if ((ticket = isam_set_ticket (file)) == -1)
			return -1;
	}

	/* 
	 * Re-located the named record, which MUST be in the database
	 */

	record2 = isam_lookup (file, index, data, IS_EQUAL, IS_UNIQUE);
	if (record != record2) {
		db_errno = DB_BAD_DB;
		goto fail;
	}
	old_block = file->f_key_block;

	if (__isam_delete (file, key, file->f_key_block,
			file->f_key_slot, file->f_key_buffer))
		goto fail;

	if (old_block == file->f_key_block)
		if (isam_balance (file, index, old_block, (keyblk_t *) 0, 0))
			goto fail;

	if (! was_locked) {
		file->f_ticket = ticket;
		isam_end_ticket (file);
	}
	return 0;

fail:
	if (! was_locked)
		isam_end_ticket (file);

	return -1;
}

/*
 * isam_delete_file - delete the named file
 *
 * isam_delete_file() uses the DBPATH variable to delete a named
 * ISAM file.
 */

isam_delete_file (name)
char	*name;
{
	char	*cp;
	char	*paths;
	char	new_name[BUFSIZ];
	char	key[BUFSIZ];
	char	data[BUFSIZ];

	/*
	 * Check the file name and get the search path (or make one up)
	 */

	if (strlen (name) > (BUFSIZ-5))
		return (db_errno = DB_INVALID), -1;

	if ((paths = getenv ("DBPATH")) == (char *) 0)
		paths = ".";

	/*
	 * Remove an already existing database file using
	 * DBPATH to locate the data.
	 */

	do {
		if (cp = strchr (paths, ':')) {
			strncpy (new_name, paths, cp - paths);
			new_name[cp - paths] = '\0';
			cp++;
		} else
			strcpy (new_name, paths);

		strcat (new_name, "/");
		strcat (new_name, name);
		strcat (new_name, ".dat");
	} while (access (new_name, 0) && (paths = cp));

	/*
	 * See if a valid directory prefix was found for the requested
	 * file.  If not, the file wasn't found and we can't delete it.
	 */

	if (paths) {
		if (cp) {
			strncpy (new_name, paths, cp - paths);
			new_name[cp - paths - 1] = '\0';
		} else
			strcpy (new_name, paths);

		strcat (new_name, "/");
		strcat (new_name, name);
		name = new_name;
	} else
		return (db_errno = DB_NOT_FOUND), -1;

	/*
	 * Figure out the names of the key and data files and remove
	 * both of them.
	 */

	sprintf (key, "%s.idx", name);
	sprintf (data, "%s.dat", name);

	if (unlink (key) || unlink (data)) {
		(void) unlink (data);	/* in case first unlink failed ... */
		return (db_errno = DB_EIO), -1;
	}
	return 0;
}
